import os

import numpy as np
import pytest

from effective_horizon.scripts.construct_tabular_policy import (
    ex as construct_tabular_policy_ex,
)
from effective_horizon.scripts.convert_atari_head_data import (
    ex as convert_atari_head_data_ex,
)
from effective_horizon.scripts.convert_atari_mdp_to_framestack import (
    ex as convert_atari_mdp_to_framestack_ex,
)


def test_convert_atari_head_data_one_step(tmp_path):
    convert_atari_head_data_ex.run(
        config_updates={
            "data_dir": "data/atari-head/freeway",
            "out_path": str(tmp_path / "json"),
            "out_format": "json",
            "rom": "freeway",
            "minimal_actions": True,
        }
    )


@pytest.mark.uses_rllib
def test_convert_atari_head_data_two_steps(tmp_path):
    from effective_horizon.scripts.filter_to_minimal_actions import (
        ex as filter_to_minimal_actions_ex,
    )

    convert_atari_head_data_ex.run(
        config_updates={
            "data_dir": "data/atari-head/freeway",
            "out_path": str(tmp_path / "rllib_complete"),
            "out_format": "rllib",
        }
    )
    filter_to_minimal_actions_ex.run(
        config_updates={
            "in_dir": str(tmp_path / "rllib_complete"),
            "out_dir": str(tmp_path / "rllib_complete_minimal_actions"),
            "rom": "freeway",
        }
    )


def test_convert_atari_mdp_to_framestack(tmp_path):
    convert_atari_mdp_to_framestack_ex.run(
        config_updates={
            "mdp": "data/mdps_with_exploration_policy_sb3/freeway_10_fs30/consolidated.npz",
            "horizon": 10,
            "out": str(tmp_path / "consolidated_framestack.npz"),
        }
    )

    framestack_mdp = np.load(tmp_path / "consolidated_framestack.npz")
    expected_framestack_mdp = np.load(
        "data/mdps_with_exploration_policy_sb3/freeway_10_fs30/consolidated_framestack.npz"
    )
    for key in framestack_mdp:
        np.testing.assert_array_equal(framestack_mdp[key], expected_framestack_mdp[key])


@pytest.mark.uses_rllib
@pytest.mark.uses_cuda
def test_construct_tabular_policy_rllib(tmp_path):
    os.makedirs(os.path.expanduser("~/ray_results/BCTrainer_mdps"), exist_ok=True)

    for mdp_fname, checkpoint_fname, run, horizon in [
        (
            "data/mdps_with_exploration_policy_rllib/freeway_10_fs30/consolidated_framestack.npz",
            "data/logs/bc/mdps/freeway_200_fs30-v0/complete/entropy_0.1/2022-12-30_16-10-41/checkpoint_000100/checkpoint-100",
            "BC",
            10,
        ),
        (
            "data/mdps_with_exploration_policy_rllib/maze_easy_l0_30_fs1/consolidated.npz",
            "data/logs/PPO/procgen/maze/entropy_0.1/2022-12-16_07-33-02/checkpoint_002500/checkpoint-2500",
            "PPO",
            30,
        ),
    ]:
        mdp_name = mdp_fname.split("/")[-2]
        out_dir = tmp_path / mdp_name
        os.makedirs(out_dir, exist_ok=True)
        construct_tabular_policy_ex.run(
            config_updates={
                "rllib_checkpoint": checkpoint_fname,
                "mdp": mdp_fname,
                "horizon": horizon,
                "run": run,
                "out": str(out_dir / "exploration_policy.npy"),
            }
        )

        exploration_policy = np.load(out_dir / "exploration_policy.npy")
        expected_exploration_policy = np.load(
            f"{os.path.dirname(mdp_fname)}/exploration_policy.npy"
        )
        np.testing.assert_array_almost_equal(
            exploration_policy,
            expected_exploration_policy,
            decimal=3,
        )


@pytest.mark.uses_sb3
@pytest.mark.uses_cuda
def test_construct_tabular_policy_sb3(tmp_path):
    for mdp_fname, policy_fname, horizon in [
        (
            "data/mdps_with_exploration_policy_sb3/freeway_10_fs30/consolidated_framestack.npz",
            "data/logs_sb3/BC/BRIDGE/freeway_200_fs30-v0/CnnPolicy/entropy_0.1/2023-10-18_23-17-48/1/final_policy.pt",
            10,
        ),
        (
            "data/mdps_with_exploration_policy_sb3/maze_easy_l0_30_fs1/consolidated.npz",
            "data/logs_sb3/PPO/procgen/maze/easy/CnnPolicy/entropy_0.1/2023-10-18_17-32-12/1/final_policy.pt",
            30,
        ),
    ]:
        mdp_name = mdp_fname.split("/")[-2]
        print(mdp_name)
        out_dir = tmp_path / mdp_name
        os.makedirs(out_dir, exist_ok=True)
        construct_tabular_policy_ex.run(
            config_updates={
                "sb3_policy": policy_fname,
                "mdp": mdp_fname,
                "horizon": horizon,
                "out": str(out_dir / "exploration_policy.npy"),
            }
        )

        exploration_policy = np.load(out_dir / "exploration_policy.npy")
        expected_exploration_policy = np.load(
            f"{os.path.dirname(mdp_fname)}/exploration_policy.npy"
        )
        np.testing.assert_array_almost_equal(
            exploration_policy,
            expected_exploration_policy,
            decimal=3,
        )



import random
import time

import pytest

from effective_horizon.envs.atari import AtariEnv
from effective_horizon.envs.deterministic_registration import (
    GYM_NAMESPACE,
    register_all,
)
from effective_horizon.envs.procgen import ProcgenEnv

try:
    from effective_horizon.rllib.train import ex as train_ex
    from effective_horizon.rllib.train_bc import ex as train_bc_ex
except ImportError:
    pass


# For some reason, pytest and the Ray registry don't play nicely together so we
# have to re-register all the environments before each test.
@pytest.fixture(autouse=True)
def register_envs():
    register_all()
    try:
        from ray.tune.registry import register_env
    except ImportError:
        pass
    else:
        register_env("procgen", ProcgenEnv)
        register_env("atari", AtariEnv)


@pytest.fixture
def random_seed():
    random.seed(time.time_ns())


@pytest.mark.uses_rllib
def test_pretrain_atari_bc(random_seed, tmp_path):
    run = train_bc_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": f"{GYM_NAMESPACE}/freeway_10_fs30-v0",
            "input": "data/atari_human_data/rllib_complete_minimal_actions/freeway",
            "num_workers": 0,
            "entropy_coeff": 0.1,
            "num_training_iters": 10,
            "train_batch_size": 100,
        }
    )
    assert run.result is not None
    assert (
        run.result["info"]["learner"]["default_policy"]["learner_stats"]["bc_loss"] < 1
    )


@pytest.mark.uses_rllib
def test_pretrain_procgen(random_seed, tmp_path):
    run = train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": "procgen",
            "procgen_env_name": "coinrun",
            "run": "PPO",
            "train_batch_size": 100,
            "rollout_fragment_length": 100,
            "sgd_minibatch_size": 100,
            "num_workers": 1,
            "num_sgd_iter": 1,
            "num_training_iters": 1,
            "entropy_coeff": 0.1,
        }
    )
    assert run.result is not None
    assert run.result["episode_reward_mean"] >= 0


@pytest.mark.uses_rllib
@pytest.mark.slow
def test_train_ppo(random_seed, tmp_path):
    mdp, optimal_reward = random.choice(
        [
            (f"{GYM_NAMESPACE}/freeway_10_fs30-v0", 1),
            (f"{GYM_NAMESPACE}/maze_easy_l0_30_fs1-v0", 10),
            (f"{GYM_NAMESPACE}/MiniGrid-KeyCorridorS3R1-v0", 1),
        ]
    )
    run = train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": mdp,
            "run": "PPO",
            "train_batch_size": 100,
            "rollout_fragment_length": 100,
            "sgd_minibatch_size": 100,
            "num_sgd_iter": 1,
            "num_workers": 1,
            "num_training_iters": 1,
            "stop_on_eval_reward": optimal_reward,
        }
    )
    assert run.result is not None
    assert run.result["episode_reward_mean"] >= 0


@pytest.mark.uses_rllib
@pytest.mark.slow
def test_train_dqn(random_seed, tmp_path):
    mdp, optimal_reward = random.choice(
        [
            (f"{GYM_NAMESPACE}/freeway_10_fs30-v0", 1),
            (f"{GYM_NAMESPACE}/maze_easy_l0_30_fs1-v0", 10),
            (f"{GYM_NAMESPACE}/MiniGrid-KeyCorridorS3R1-v0", 1),
        ]
    )
    run = train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": mdp,
            "run": "FastDQN",
            "train_batch_size": 80,
            "rollout_fragment_length": 10,
            "sgd_minibatch_size": 80,
            "num_sgd_iter": 1,
            "num_workers": 1,
            "epsilon_timesteps": 1000,
            "dueling": True,
            "double_q": True,
            "prioritized_replay": True,
            "replay_buffer_capacity": 1000,
            "learning_starts": 0,
            "num_training_iters": 1,
            "stop_on_eval_reward": optimal_reward,
        }
    )
    assert run.result is not None
    assert run.result["episode_reward_mean"] >= 0


@pytest.mark.uses_rllib
def test_train_gorp(random_seed, tmp_path):
    run = train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": "atari",
            "horizon": 27000,
            "frameskip": 4,
            "deterministic": True,
            "rom_file": "pong",
            "reward_scale": 1,
            "done_on_life_lost": True,
            "gamma": 0.99,
            "run": "GORP",
            "episodes_per_action_seq": 1,
            "num_training_iters": 2,
            "stop_on_timesteps": 1000,
        }
    )
    assert run.result is not None
    assert run.result["episode_reward_mean"] >= -21



import glob
import logging
import os
import random
import re
from typing import Any, List, cast

import gymnasium as gym
import numpy as np

from effective_horizon.envs.deterministic_registration import GYM_NAMESPACE
from effective_horizon.mdp_utils import get_sparse_mdp, load_mdp, run_value_iteration

logger = logging.getLogger(__name__)


def test_random_deterministic_mdp():
    mdp_fnames = glob.glob("data/mdps/*/consolidated.npz")
    mdp_fnames.sort()
    mdp_fname = random.choice(mdp_fnames)

    mdp_name = os.path.basename(os.path.dirname(mdp_fname))

    logger.info(f"loading {mdp_fname}...")
    transitions, rewards = load_mdp(mdp_fname)
    num_states, num_actions = transitions.shape
    done_state = num_states - 1

    # Sometimes, set the goal to just reach a done state.
    rewards_for_vi = rewards.copy()
    if random.random() < 0.1:
        rewards_for_vi[:, :] = 0
        rewards_for_vi[transitions == done_state] = 1
        rewards_for_vi[done_state, :] = 0

    sparse_transitions, rewards_vector = get_sparse_mdp(transitions, rewards_for_vi)

    mdp_name = os.path.basename(os.path.dirname(mdp_fname))
    env_id = f"{GYM_NAMESPACE}/{mdp_name}"
    if not re.match(r".*-v\d+", env_id):
        env_id += "-v0"
    logger.info(f"creating gym env {env_id}")
    env = gym.make(env_id)
    horizon: int = cast(Any, env).horizon
    reward_scale = getattr(env, "reward_scale", 1)

    if num_states * horizon >= 100_000_000:
        return

    logger.info("running value iteration")
    vi = run_value_iteration(sparse_transitions, rewards_vector, horizon)

    num_episodes = 100
    logger.info(f"running {num_episodes} episodes")
    for _ in range(num_episodes):
        if random.random() < 0.2:
            guide_qs = -vi.worst_qs
        else:
            guide_qs = vi.optimal_qs
        eps = random.random()

        state = 0
        actions: List[int] = []
        env.reset()
        total_reward = 0
        for t in range(horizon):
            if random.random() < eps:
                action = random.randrange(num_actions)
            else:
                state_qs = guide_qs[t, state]
                action = random.choice(
                    np.arange(num_actions)[state_qs == state_qs.max()]
                )
            actions.append(action)

            expected_reward = rewards[state, action] * reward_scale
            _, env_reward, terminated, truncated, _ = env.step(action)
            done = terminated or truncated
            if abs(env_reward - expected_reward) > 1e-4:
                raise ValueError(
                    f"expected reward {expected_reward} but received {env_reward} "
                    f"from following actions {actions}"
                )
            total_reward += expected_reward

            state = int(transitions[state, action])
            if done:
                if "MiniGrid" in env_id and t == horizon - 1:
                    # MiniGrid envs don't terminate on horizon.
                    pass
                elif not (
                    state == done_state or np.all(transitions[state] == done_state)
                ):
                    raise ValueError(
                        f"expected done state {done_state} but reached {state} "
                        f"with transitions {transitions[state].astype(int)} "
                        f"from following actions {actions}"
                    )
                break
            elif state == done_state:
                raise ValueError(
                    f"expected none-done state but reached done state "
                    f"with transitions {transitions[state].astype(int)} "
                    f"from following actions {actions}"
                )

        logger.info(f"actions {' '.join(map(str, actions))} => reward {total_reward}")


def test_atari_framestack():
    mdp_fnames = glob.glob(
        "data/mdps_with_exploration_policy_sb3/*/consolidated_framestack.npz"
    )
    mdp_fnames.sort()
    mdp_fname = random.choice(mdp_fnames)

    mdp_name = os.path.basename(os.path.dirname(mdp_fname))

    logger.info(f"loading {mdp_fname}...")
    transitions, rewards = load_mdp(mdp_fname)
    num_states, num_actions = transitions.shape
    done_state = num_states - 1
    mdp = np.load(mdp_fname)
    screens = mdp["screens"]
    screen_ids = mdp["screen_mapping"]

    mdp_name = os.path.basename(os.path.dirname(mdp_fname))
    env_id = f"{GYM_NAMESPACE}/{mdp_name}-v0"
    logger.info(f"creating gym env {env_id}")
    env = gym.make(env_id)
    horizon: int = cast(Any, env).horizon
    reward_scale: float = cast(Any, env).reward_scale

    num_episodes = 100
    logger.info(f"running {num_episodes} episodes")
    for _ in range(num_episodes):
        state = 0
        actions: List[int] = []
        obs, _ = env.reset()
        for t in range(horizon):
            action = random.randrange(num_actions)
            actions.append(action)

            expected_screen = screens[screen_ids[state]].transpose(1, 2, 0)
            np.testing.assert_allclose(
                expected_screen,
                obs,
                rtol=0,
                atol=0.6 / 255,
                err_msg=f"mismatched screens from following actions {actions}",
            )

            expected_reward = rewards[state, action] * reward_scale
            obs, env_reward, terminated, truncated, _ = env.step(action)
            done = terminated or truncated
            if env_reward != expected_reward:
                raise ValueError(
                    f"expected reward {expected_reward} but received {env_reward} "
                    f"from following actions {actions}"
                )

            state = int(transitions[state, action])
            if done:
                if not (
                    state == done_state or np.all(transitions[state] == done_state)
                ):
                    raise ValueError(
                        f"expected done state {done_state} but reached {state} "
                        f"with transitions {transitions[state].astype(int)} "
                        f"from following actions {actions}"
                    )
                break

        logger.info(f"actions {' '.join(map(str, actions))}")



import json
import os
import random
import subprocess

import numpy as np
import pytest


def assert_structure_close(a, b, eps=1e-8):
    if isinstance(a, dict):
        assert isinstance(b, dict)
        assert set(a.keys()) == set(b.keys())
        for k in a:
            assert_structure_close(a[k], b[k], eps=eps)
    elif isinstance(a, list):
        assert isinstance(b, list)
        assert len(a) == len(b)
        for a_i, b_i in zip(a, b):
            assert_structure_close(a_i, b_i, eps=eps)
    elif isinstance(a, np.ndarray):
        assert isinstance(b, np.ndarray)
        assert a.shape == b.shape
        assert np.all(np.abs(a - b) < eps)
    else:
        assert (a == b) or abs(a - b) < eps


TEST_MDPS = [
    ("data/mdps/MiniGrid-KeyCorridorS3R1-v0/consolidated.npz", 100),
    ("data/mdps/freeway_10_fs30/consolidated.npz", 10),
    ("data/mdps/maze_easy_l0_30_fs1/consolidated.npz", 30),
    (
        "data/mdps_with_exploration_policy_sb3/freeway_10_fs30/consolidated_framestack.npz",
        10,
    ),
    ("data/mdps_with_exploration_policy_sb3/maze_easy_l0_30_fs1/consolidated.npz", 30),
]


@pytest.mark.uses_julia
def test_construct_atari_mdp(tmp_path):
    subprocess.check_call(
        [
            "julia",
            "--threads=2",
            "--project=EffectiveHorizon.jl",
            "EffectiveHorizon.jl/src/construct_mdp.jl",
            "--rom",
            "freeway",
            "-o",
            tmp_path,
            "--horizon",
            "10",
            "--frameskip",
            "30",
            "--done_on_life_lost",
            "--save_screens",
            "--save_states",
        ]
    )
    for mdp_fname, expected_num_states in [
        ("mdp.npz", 399),
        ("consolidated.npz", 197),
        ("consolidated_ignore_screen.npz", 33),
    ]:
        mdp = np.load(tmp_path / mdp_fname)
        assert mdp["transitions"].shape == (expected_num_states, 3)
        assert mdp["screen_mapping"].shape == (expected_num_states,)
        state_lengths = mdp["state_lengths"]
        assert state_lengths.shape == (expected_num_states,)
        assert mdp["states"].shape == (expected_num_states, state_lengths.max())


@pytest.mark.uses_julia
def test_construct_procgen_mdp(tmp_path):
    subprocess.check_call(
        [
            "julia",
            "--threads=2",
            "--project=EffectiveHorizon.jl",
            "EffectiveHorizon.jl/src/construct_mdp.jl",
            "--env_name",
            "maze",
            "--distribution_mode",
            "easy",
            "--level",
            "0",
            "-o",
            tmp_path,
            "--horizon",
            "30",
            "--frameskip",
            "1",
            "--save_screens",
            "--save_states",
        ]
    )
    for mdp_fname, expected_num_states in [
        ("mdp.npz", 2227),
        ("consolidated.npz", 243),
        ("consolidated_ignore_screen.npz", 134),
    ]:
        mdp = np.load(tmp_path / mdp_fname)
        assert mdp["transitions"].shape == (expected_num_states, 9)
        assert mdp["screen_mapping"].shape == (expected_num_states,)
        state_lengths = mdp["state_lengths"]
        assert state_lengths.shape == (expected_num_states,)
        assert mdp["states"].shape == (expected_num_states, state_lengths.max())


@pytest.mark.uses_julia
def test_construct_minigrid_mdp(tmp_path):
    subprocess.check_call(
        [
            "julia",
            "--project=EffectiveHorizon.jl",
            "EffectiveHorizon.jl/src/construct_mdp.jl",
            "--minigrid",
            "--env_name",
            "BRIDGE/MiniGrid-KeyCorridorS3R1-v0",
            "-o",
            tmp_path,
            "--horizon",
            "100000000",
            "--frameskip",
            "1",
            "--save_screens",
            "--save_states",
        ]
    )
    for mdp_fname, expected_num_states in [
        ("mdp.npz", 169),
        ("consolidated.npz", 168),
        ("consolidated_ignore_screen.npz", 168),
    ]:
        mdp = np.load(tmp_path / mdp_fname)
        assert mdp["transitions"].shape == (expected_num_states, 6)
        assert mdp["screen_mapping"].shape == (expected_num_states,)
        state_lengths = mdp["state_lengths"]
        assert state_lengths.shape == (expected_num_states,)
        assert mdp["states"].shape == (expected_num_states, state_lengths.max())


@pytest.mark.uses_julia
def test_analyze_mdp(tmp_path, randomize_julia_scripts):
    test_mdps = TEST_MDPS
    if randomize_julia_scripts:
        test_mdps = random.sample(test_mdps, 1)

    for mdp_fname, horizon in test_mdps:
        mdp_type = mdp_fname.split("/")[1]
        mdp_name = mdp_fname.split("/")[-2]
        out_dir = tmp_path / mdp_type / mdp_name
        os.makedirs(out_dir, exist_ok=True)
        analyze_command = [
            "julia",
            "--project=EffectiveHorizon.jl",
            "EffectiveHorizon.jl/src/analyze_mdp.jl",
            "--mdp",
            mdp_fname,
            "--horizon",
            str(horizon),
            "-o",
            out_dir / "consolidated_analyzed.json",
        ]
        if mdp_type == "mdps_with_exploration_policy_sb3":
            analyze_command.extend(
                [
                    "--exploration_policy",
                    os.path.dirname(mdp_fname) + "/exploration_policy.npy",
                ]
            )
        subprocess.check_call(analyze_command)
        with open(out_dir / "consolidated_analyzed.json") as results_file:
            results = json.load(results_file)
        with open(mdp_fname[:-4] + "_analyzed.json") as expected_results_file:
            expected_results = json.load(expected_results_file)
        assert_structure_close(results, expected_results)

        value_dists = np.load(out_dir / "consolidated_analyzed_value_dists_1.npy")
        expected_value_dists = np.load(mdp_fname[:-4] + "_analyzed_value_dists_1.npy")
        assert_structure_close(value_dists, expected_value_dists)


@pytest.mark.uses_julia
def test_analyze_sticky_actions(tmp_path, randomize_julia_scripts):
    test_mdps = TEST_MDPS
    if randomize_julia_scripts:
        test_mdps = random.sample(test_mdps, 1)

    for mdp_fname, horizon in test_mdps:
        mdp_type = mdp_fname.split("/")[1]
        if mdp_type == "mdps_with_exploration_policy_sb3":
            # We don't analyze exploration policies for sticky actions.
            continue

        mdp_name = mdp_fname.split("/")[-2]
        out_dir = tmp_path / mdp_type / mdp_name
        os.makedirs(out_dir, exist_ok=True)
        analyze_command = [
            "julia",
            "--project=EffectiveHorizon.jl",
            "EffectiveHorizon.jl/src/analyze_sticky_actions.jl",
            "--mdp",
            mdp_fname,
            "--horizon",
            str(horizon),
            "-o",
            out_dir / "consolidated_analyzed_sticky_0.25.json",
        ]
        subprocess.check_call(analyze_command)
        with open(out_dir / "consolidated_analyzed_sticky_0.25.json") as results_file:
            results = json.load(results_file)
        with open(
            mdp_fname[:-4] + "_analyzed_sticky_0.25.json"
        ) as expected_results_file:
            expected_results = json.load(expected_results_file)
        assert_structure_close(results, expected_results)


@pytest.mark.uses_julia
def test_compute_gorp_bounds(tmp_path, randomize_julia_scripts):
    test_mdps = TEST_MDPS
    if randomize_julia_scripts:
        test_mdps = random.sample(test_mdps, 1)

    for mdp_fname, horizon in test_mdps:
        # MiniGrid environments take forever because of the long horizon and
        # because almost all states are optimal, so we skip them.
        if "MiniGrid" in mdp_fname:
            continue

        mdp_type = mdp_fname.split("/")[1]
        mdp_name = mdp_fname.split("/")[-2]
        out_dir = tmp_path / mdp_type / mdp_name
        os.makedirs(out_dir, exist_ok=True)

        with open(mdp_fname[:-4] + "_gorp_bounds.json") as expected_results_file:
            expected_results = json.load(expected_results_file)
        max_k = len(expected_results["k_results"])

        gorp_bounds_command = [
            "julia",
            "--threads=2",
            "--project=EffectiveHorizon.jl",
            "EffectiveHorizon.jl/src/compute_gorp_bounds.jl",
            "--mdp",
            mdp_fname,
            "--use_value_dists",
            "--horizon",
            str(horizon),
            "--max_k",
            str(max_k),
            "-o",
            out_dir / "consolidated_gorp_bounds.json",
        ]
        if mdp_type == "mdps_with_exploration_policy_sb3":
            gorp_bounds_command.extend(
                [
                    "--exploration_policy",
                    os.path.dirname(mdp_fname) + "/exploration_policy.npy",
                ]
            )
        subprocess.check_call(gorp_bounds_command)

        with open(out_dir / "consolidated_gorp_bounds.json") as results_file:
            results = json.load(results_file)
        assert_structure_close(results, expected_results)


@pytest.mark.uses_julia
def test_run_gorp_and_ucb(tmp_path):
    for mdp_fname, horizon in TEST_MDPS:
        mdp_type = mdp_fname.split("/")[1]
        mdp_name = mdp_fname.split("/")[-2]
        out_dir = tmp_path / mdp_type / mdp_name
        os.makedirs(out_dir, exist_ok=True)

        value_dists = np.load(mdp_fname[:-4] + "_analyzed_value_dists_1.npy")
        initial_value_dist = value_dists[:, :, 0]
        optimal_return = initial_value_dist[0].max()

        alg = random.choice(["ucb", "gorp"])

        command = [
            "julia",
            "--threads=2",
            "--project=EffectiveHorizon.jl",
            f"EffectiveHorizon.jl/src/run_{alg}.jl",
            "--mdp",
            mdp_fname,
            "--horizon",
            str(horizon),
            "--max_sample_complexity",
            "1000000",
            "--num_runs",
            "5",
            "-o",
            out_dir / f"consolidated_{alg}.json",
            "--optimal_return",
            str(optimal_return),
        ]
        if alg == "gorp":
            k = np.random.randint(1, 3)
            command.extend(["--k", str(k)])
        if mdp_type == "mdps_with_exploration_policy_sb3" and alg == "gorp":
            command.extend(
                [
                    "--exploration_policy",
                    os.path.dirname(mdp_fname) + "/exploration_policy.npy",
                ]
            )
        subprocess.check_call(command)

        with open(out_dir / f"consolidated_{alg}.json") as results_file:
            results = json.load(results_file)

        if alg == "gorp":
            assert results.keys() == {
                "actions",
                "mean_rewards",
                "final_returns",
                "sample_complexities",
                "median_final_return",
                "median_sample_complexity",
                "timesteps_per_iteration",
            }
        else:
            assert results.keys() == {
                "actions",
                "final_returns",
                "sample_complexities",
                "median_final_return",
                "median_sample_complexity",
            }



import os
import random
from typing import List, Tuple

import pytest

from effective_horizon.envs.deterministic_registration import (
    GYM_NAMESPACE,
    register_all,
)

try:
    from effective_horizon.sb3.train import ex as train_ex
except ImportError:
    pass


# For some reason, pytest and the gym registry don't play nicely together so we
# have to re-register all the environments before each test.
@pytest.fixture(autouse=True)
def register_envs():
    register_all()


mdps_for_testing: List[Tuple[str, int]] = [
    (f"{GYM_NAMESPACE}/freeway_10_fs30-v0", 1),
    (f"{GYM_NAMESPACE}/maze_easy_l2_30_fs1-v0", 10),
    (f"{GYM_NAMESPACE}/MiniGrid-Empty-5x5-v0", 1),
]


@pytest.mark.uses_sb3
def test_train_ppo(tmp_path):
    mdp, optimal_reward = random.choice(mdps_for_testing)
    train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": mdp,
            "algo": "PPO",
            "timesteps": 4_000,
            "stop_on_eval_reward": optimal_reward,
        }
    )


@pytest.mark.uses_sb3
def test_train_dqn(tmp_path):
    mdp, optimal_reward = random.choice(mdps_for_testing)
    train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": mdp,
            "algo": "DQN",
            "timesteps": 4_000,
            "algo_args": {"exploration_fraction": 1.0, "learning_starts": 0},
            "stop_on_eval_reward": optimal_reward,
        }
    )


@pytest.mark.uses_sb3
def test_train_sqirl(tmp_path):
    train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": "BRIDGE/Atari-v0",
            "rom_file": "breakout",
            "done_on_life_lost": True,
            "fire_on_reset": True,
            "horizon": 27_000,
            "frameskip": 4,
            "gamma": 0.99,
            "deterministic": False,
            "algo": "SQIRL",
            "use_impala_cnn": True,
            "timesteps": 4_000,
            "algo_args": {
                "n_epochs": 1,
                "episodes_per_timestep": 1,
                "planning_depth": 2,
                "buffer_size": 10_000,
            },
        }
    )


@pytest.mark.uses_sb3
def test_train_gorp(tmp_path):
    train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": "BRIDGE/Procgen-v0",
            "procgen_env_name": "coinrun",
            "start_level": 0,
            "num_levels": 1,
            "rand_seed": 0,
            "algo": "GORP",
            "timesteps": 10_000,
            "algo_args": {
                "episodes_per_action_sequence": 1,
                "planning_depth": 2,
            },
        }
    )


@pytest.mark.uses_sb3
def test_pretrain_atari_bc(tmp_path):
    env_name = f"{GYM_NAMESPACE}/freeway_10_fs30-v0"

    ex = train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": env_name,
            "bc_data_fname": "data/atari_human_data/json/freeway.json",
            "bc_epochs": 1,
            "algo": "BC",
            "algo_args": {
                "ent_weight": 0.1,
                "batch_size": 4,
                "minibatch_size": 4,
            },
        }
    )

    assert ex.result is not None
    policy_path = ex.result["policy_path"]
    assert os.path.exists(policy_path)

    train_ex.run(
        config_updates={
            "log_dir": tmp_path / "logs",
            "env_name": env_name,
            "algo": "PPO",
            "timesteps": 10_000,
            "stop_on_eval_reward": 1,
            "pretrained_policy_fname": policy_path,
        }
    )
